@@ -30,13 +30,20 @@ static void iommu_dump_p2m_table(unsigned char key);
  *
  *   off|no|false|disable       Disable IOMMU (default)
  *   force|required             Don't boot unless IOMMU is enabled
+ *   no-intremap                Disable interrupt remapping
+ *   intpost                    Enable VT-d Interrupt posting
+ *   verbose                    Be more verbose
+ *   debug                      Enable debugging messages and checks
  *   workaround_bios_bug        Workaround some bios issue to still enable
-                                VT-d, don't guarantee security
+ *                              VT-d, don't guarantee security
  *   dom0-passthrough           No DMA translation at all for Dom0
  *   dom0-strict                No 1:1 memory mapping for Dom0
+ *   no-sharept                 Don't share VT-d and EPT page tables
  *   no-snoop                   Disable VT-d Snoop Control
  *   no-qinval                  Disable VT-d Queued Invalidation
- *   no-intremap                Disable VT-d Interrupt Remapping
+ *   no-igfx                    Disable VT-d for IGD devices (insecure)
+ *   no-amd-iommu-perdev-intremap Don't use per-device interrupt remapping
+ *                              tables (insecure)
  */
 custom_param("iommu", parse_iommu_param);
 bool_t __initdata iommu_enable = 1;
@@ -50,6 +57,14 @@ bool_t __read_mostly iommu_passthrough;
 bool_t __read_mostly iommu_snoop = 1;
 bool_t __read_mostly iommu_qinval = 1;
 bool_t __read_mostly iommu_intremap = 1;
+
+/*
+ * In the current implementation of VT-d posted interrupts, in some extreme
+ * cases, the per cpu list which saves the blocked vCPU will be very long,
+ * and this will affect the interrupt latency, so let this feature off by
+ * default until we find a good solution to resolve it.
+ */
+bool_t __read_mostly iommu_intpost;
 bool_t __read_mostly iommu_hap_pt_share = 1;
 bool_t __read_mostly iommu_debug;
 bool_t __read_mostly amd_iommu_perdev_intremap = 1;
@@ -60,12 +75,6 @@ DEFINE_SPINLOCK(iommu_pt_cleanup_lock);
 PAGE_LIST_HEAD(iommu_pt_cleanup_list);
 static struct tasklet iommu_pt_cleanup_tasklet;
 
-static struct keyhandler iommu_p2m_table = {
-    .diagnostic = 0,
-    .u.fn = iommu_dump_p2m_table,
-    .desc = "dump iommu p2m table"
-};
-
 static void __init parse_iommu_param(char *s)
 {
     char *ss;
@@ -96,6 +105,8 @@ static void __init parse_iommu_param(char *s)
             iommu_qinval = val;
         else if ( !strcmp(s, "intremap") )
             iommu_intremap = val;
+        else if ( !strcmp(s, "intpost") )
+            iommu_intpost = val;
         else if ( !strcmp(s, "debug") )
         {
             iommu_debug = val;
@@ -154,7 +165,7 @@ void __hwdom_init iommu_hwdom_init(struct domain *d)
     if ( !iommu_enabled )
         return;
 
-    register_keyhandler('o', &iommu_p2m_table);
+    register_keyhandler('o', &iommu_dump_p2m_table, "dump iommu p2m table", 0);
     d->need_iommu = !!iommu_dom0_strict;
     if ( need_iommu(d) && !iommu_use_hap_pt(d) )
     {
@@ -219,7 +230,8 @@ void iommu_domain_destroy(struct domain *d)
     if ( !iommu_enabled || !dom_iommu(d)->platform_ops )
         return;
 
-    iommu_teardown(d);
+    if ( need_iommu(d) )
+        iommu_teardown(d);
 
     arch_iommu_domain_destroy(d);
 }
@@ -303,6 +315,9 @@ int __init iommu_setup(void)
         panic("Couldn't enable %s and iommu=required/force",
               !iommu_enabled ? "IOMMU" : "Interrupt Remapping");
 
+    if ( !iommu_intremap )
+        iommu_intpost = 0;
+
     if ( !iommu_enabled )
     {
         iommu_snoop = 0;
@@ -322,11 +337,16 @@ int __init iommu_setup(void)
     return rc;
 }
 
+void iommu_suspend()
+{
+    if ( iommu_enabled )
+        iommu_get_ops()->suspend();
+}
+
 void iommu_resume()
 {
-    const struct iommu_ops *ops = iommu_get_ops();
     if ( iommu_enabled )
-        ops->resume();
+        iommu_get_ops()->resume();
 }
 
 int iommu_do_domctl(
@@ -338,11 +358,11 @@ int iommu_do_domctl(
     if ( !iommu_enabled )
         return -ENOSYS;
 
-#ifdef HAS_PCI
+#ifdef CONFIG_HAS_PCI
     ret = iommu_do_pci_domctl(domctl, d, u_domctl);
 #endif
 
-#ifdef HAS_DEVICE_TREE
+#ifdef CONFIG_HAS_DEVICE_TREE
     if ( ret == -ENODEV )
         ret = iommu_do_dt_domctl(domctl, d, u_domctl);
 #endif
@@ -350,34 +370,28 @@ int iommu_do_domctl(
     return ret;
 }
 
-void iommu_suspend()
-{
-    const struct iommu_ops *ops = iommu_get_ops();
-    if ( iommu_enabled )
-        ops->suspend();
-}
-
 void iommu_share_p2m_table(struct domain* d)
 {
-    const struct iommu_ops *ops = iommu_get_ops();
-
     if ( iommu_enabled && iommu_use_hap_pt(d) )
-        ops->share_p2m(d);
+        iommu_get_ops()->share_p2m(d);
 }
 
 void iommu_crash_shutdown(void)
 {
-    const struct iommu_ops *ops = iommu_get_ops();
     if ( iommu_enabled )
-        ops->crash_shutdown();
-    iommu_enabled = iommu_intremap = 0;
+        iommu_get_ops()->crash_shutdown();
+    iommu_enabled = iommu_intremap = iommu_intpost = 0;
 }
 
 int iommu_get_reserved_device_memory(iommu_grdm_t *func, void *ctxt)
 {
-    const struct iommu_ops *ops = iommu_get_ops();
+    const struct iommu_ops *ops;
 
-    if ( !iommu_enabled || !ops->get_reserved_device_memory )
+    if ( !iommu_enabled )
+        return 0;
+
+    ops = iommu_get_ops();
+    if ( !ops->get_reserved_device_memory )
         return 0;
 
     return ops->get_reserved_device_memory(func, ctxt);
